#!/usr/bin/env python3
import os
from hashlib import md5
from datetime import datetime
from contextlib import contextmanager

from itertools import chain

from bsddb3.db import DB
from bsddb3.db import DBEnv
from bsddb3.db import DB_HASH
from bsddb3.db import DB_CREATE
from bsddb3.db import DB_INIT_LOG
from bsddb3.db import DB_INIT_TXN
from bsddb3.db import DB_INIT_MPOOL
from bsddb3.db import DB_MULTIVERSION
from bsddb3.db import DB_RECOVER_FATAL
from bsddb3.db import *

from msgpack import dumps as msgpack_dumps
from msgpack import loads as msgpack_loads


def dumps(data):
    return msgpack_dumps(data, encoding='utf-8')


def loads(data):
    return msgpack_loads(data, encoding='utf-8')


__version__ = '14.10.04-dev'


VERTEX, EDGE = range(2)


class AjguException(Exception):
    pass


class Element:
    """Base class for Vertex and Edge."""

    # there should be a __init__ method

    def __getattr__(self, key):
        """Implement `magic.attribute`"""
        try:
            return self[key]
        except KeyError:
            raise AttributeError('no `{}` method or property key'.format(key))

    def identifier(self):
        """Returns a string identifier of the element"""
        return self._identifier

    def label(self):
        """Return the string label associated with thi label"""
        return self._label

    def properties(self):
        """Returns the dictionary of properties of this element"""
        return self._properties

    def __setitem__(self, key, value):
        """set property `key` to `value`"""
        self._txn._dirty.add(self)
        try:
            self._properties[key]
        except KeyError:
            self._properties[key] = value
        else:
            try:
                self._original_properties[key]
            except KeyError:
                original_value = self._properties[key]
                self._original_properties[key] = original_value
            finally:
                self._properties[key] = value

    def __getitem__(self, key):
        """get property `key` of this element"""
        return self._properties[key]

    def get(self, key, default):
        """get property `key` returns `default` if there is none"""
        return self._properties.get(key, default)

    def _save(self):
        """pack element and save it the database

        :internal"""
        values = self._values()
        values = dumps(values)
        identifier = self._identifier.encode('ascii')
        self._txn._database.elements.put(identifier, values)

    def original(self, key):
        """Returns original value associated with `key`

        "original value" means the value it had when the element was
        loaded. It can raise a `KeyError` if there is no original value."""
        return self._previous_properties[key]


class Vertex(Element):
    """Represent a vertex in graph database

    Vertex objects are created via `Transaction` objects"""

    def __init__(self, txn, identifier, label, outgoings, incomings, properties):
        self._txn = txn
        self._identifier = identifier
        self._label = label
        self._outgoings = outgoings
        self._incomings = incomings
        self._properties = properties
        self._previous_properties = dict()

    def _values(self):
        """create `value` associated with `self.identifier()` the `key`
        in the key/value store

        :internal"""
        values = [
            VERTEX,
            self._label,
            self._outgoings,
            self._incomings,
            self._properties
        ]
        return values

    def values(self):
        """create `value` associated with `self.identifier()` the `key`
        in the key/value store

        :internal"""
        return dict(
            label=self._label,
            outgoings=self._outgoings,
            incomings=self._incomings,
            properties=self._properties
        )

    @classmethod
    def create(cls, txn, label):
        """Called by a transaction to build a vertex.

        This is for creating a new Vertex in the database. For wrapping
        a Vertex already in the database one must use directly the
        constructor"""
        # values is built as follow:
        # element type, label, outgoings, incomings, properties
        values = [VERTEX, label, dict(), dict(), dict()]
        dump = dumps(values)
        # retrieve an identifier
        identifier = txn._random_identifier(dump)
        # store initial vertex data
        key = identifier.encode('ascii')
        txn._database.elements.put(key, dump)
        # instantiate related object
        vertex = Vertex(txn, identifier, *values[1:])
        txn._cache[identifier] = vertex
        txn._dirty.add(vertex)
        return vertex

    def _iter_over_edges(self, kind, label=None):
        """Iterate over edges of `kind` possibly filtering them by `label`

        `kind` can be `"ougoings"` or `"incomings"`.

        :internal"""
        kind = '_%s' % kind
        kind = getattr(self, kind, dict())
        if label:
            try:
                identifiers = kind[label]
            except KeyError:
                raise StopIteration
            else:
                # if identifier == 'indexed':
                #     name = '%s:%s:%s' % (self.identifier(), kind, label)
                #     index = self._txn.indexes[name]
                #     for identifier in index:
                #         yield self._txn.get(identifier)
                # else:
                #     for identifier in identifiers:
                #         yield self._txn.get(identifier)
                for identifier in identifiers:
                    yield self._txn.get(identifier)
        else:
            for identifiers in kind.values():
                # if identifiers == 'indexed':
                #     name = '%s:%s:%s' % (self.identifier(), kind, label)
                #     index = self._txn._database.indexes[name]
                #     for identifier in index:
                #         yield self._txn.get(identifier)
                # else:
                #     for identifier in identifiers:
                #         yield self._txn.get(identifier)
                for identifier in identifiers:
                    yield self._txn.get(identifier)

    def incomings(self, label=None):
        """Iterate over incoming edges possibly filtering by label

        label must be a string"""
        return self._iter_over_edges('incomings', label)

    def outgoings(self, label=None):
        """Iterate over outgoing edges possibly filtering by label

        label must be a string"""
        return self._iter_over_edges('outgoings', label)

    def delete(self):
        """Get rid of this useless node"""
        for edge in chain(self.outgoings(), self.incomings()):
            edge.delete()
        key = self._identifier.encode('ascii')
        self._txn._database.elements.delete(key)


class Edge(Element):
    """Represent an edge in graph database

    Edge objects are created via `Transaction` objects"""

    def __init__(self, txn, identifier, start, label, end, properties):
        """Constructor is only used directly by a transaction to wrap an existing
        edge. Otherwise transaction objects use `Edge.create` to construct
        an edge that must be added in the database"""
        self._txn = txn
        self._identifier = identifier
        self._start = start
        self._label = label
        self._end = end
        self._properties = properties
        self._previous_properties = dict()

    def _values(self):
        """create `value` associated with `self.identifier()` the `key`
        in the key/value store

        :internal"""
        values = [
            EDGE,
            self.start()._identifier,
            self._label,
            self.end()._identifier,
            self._properties
        ]
        return values

    def values(self):
        """create `value` associated with `self.identifier()` the `key`
        in the key/value store

        :internal"""
        return dict(
            start=self.start()._identifier,
            label=self._label,
            end=self.end()._identifier,
            properties=self._properties
        )

    @classmethod
    def create(cls, txn, start, label, end):
        """Called by a transaction to build an edge.

        This is for creating a new edge in the database. For wrapping
        an edge already in the database one must use directly the
        constructor"""
        # see also comments in Vertex.create
        values = [
            EDGE,
            start._identifier,
            label,
            end._identifier,
            dict()
        ]
        dump = dumps(values)
        identifier = txn._random_identifier(dump)
        key = identifier.encode('ascii')
        txn._database.elements.put(key, dump)
        edge = Edge(txn, identifier, *values[1:])
        txn._cache[identifier] = edge
        txn._dirty.add(edge)
        # update start and end
        try:
            start._outgoings[edge.label()].append(edge.identifier())
        except KeyError:
            start._outgoings[edge.label()] = [edge.identifier()]
        try:
            end._incomings[edge.label()].append(edge.identifier())
        except KeyError:
            end._incomings[edge.label()] = [edge.identifier()]
        # mark vertices as dirty to be updated
        txn._dirty.add(start)
        txn._dirty.add(end)
        return edge

    def start(self):
        """Return the vertex where this edge starts"""
        return self._txn.get(self._start)

    def end(self):
        """Return the vertex where this edge ends"""
        return self._txn.get(self._end)

    def label(self):
        """Return the label"""
        return self._label

    def delete(self):
        """Get rid of this useless link"""
        key = self._identifier.encode('ascii')
        self._txn._database.elements.delete(key)
        # update start and end nodes
        self.start()._outgoings[self.label()].remove(self._identifier)
        self._txn._dirty.add(self.start())
        self.end()._incomings[self.label()].remove(self._identifier)
        self._txn._dirty.add(self.end())


class Transaction:

    def __init__(self, database):
        self._txn = database.env.txn_begin()
        self._database = database
        self._dirty = set()
        self._cache = dict()
        self._indexes = set()

    def _random_identifier(self, dump):
        while True:
            now = datetime.now()
            now = now.isoformat()
            data = now.encode('ascii')
            data = data + dump
            identifier = md5(data)
            identifier = identifier.hexdigest()
            try:
                self.get(identifier)
            except AjguException:
                break
            else:
                continue
        return identifier

    def query(self, name):
        """Create a query with the given name against the current transaction"""
        from query import Query
        query = Query(self, name)
        # self._dirty.add(query)
        return query

    def index(self, name):
        index = IndexDescriptor(self, name)
        self._dirty.add(index)
        return index

    def all(self):
        cursor = self._database.elements.cursor(txn=self._txn)
        next = cursor.first()
        while next:
            identifier, dump = next
            identifier = identifier.decode('ascii')
            element = loads(dump)
            kind, values = element[0], element[1:]
            if kind == VERTEX:
                ElementClass = Vertex
            else:
                ElementClass = Edge
            element = ElementClass(self, identifier, *values)
            yield element
            next = cursor.next()

    def label(self, label):
        for element in self.all():
            if element.label() == label:
                yield element

    def get(self, identifier):
        try:
            element = self._cache[identifier]
        except KeyError:
            key = identifier.encode('ascii')
            values = self._database.elements.get(key)
            if not values:
                msg = 'Element with id=%s not found' % identifier
                raise AjguException(msg)
            values = loads(values)
            if values[0] == VERTEX:
                element = Vertex(self, identifier, *values[1:])
            else:
                element = Edge(self, identifier, *values[1:])
            self._cache[identifier] = element
        return element

    def vertex(self, label=None):
        return Vertex.create(self, label)

    def edge(self, start, label, end):
        return Edge.create(self, start, label, end)

    def commit(self):
        for element in self._dirty:
            element._save()
        self._txn.commit()
        for index in self._indexes:
            self._database.indexes.append(index)

    def rollback(self):
        self._txn.abort()
        # close indexes that were created during this transaction
        for index in self._indexes:
            index._storage.close()


class Database:

    def __init__(self, path, cache_size=(0, 512), max_cache_size=(0, 512)):
        if not os.path.exists(path):
            os.makedirs(path)
        self.env = DBEnv()
        self.env.set_cache_max(*max_cache_size)
        flags = (DB_INIT_LOG |
                 DB_INIT_TXN |
                 DB_RECOVER_FATAL |
                 DB_CREATE |
                 DB_INIT_MPOOL
        )
        self.env.open(
            path,
            flags,
            0
        )
        self.env.set_cachesize(*cache_size)

        txn = self.txn()
        flags = DB_CREATE | DB_MULTIVERSION
        self.elements = DB(self.env)
        self.elements.open('ajgu-elements', None, DB_HASH, flags, 0, txn=txn._txn)
        txn.commit()

        self.indexes = list()

    def close(self):
        self.elements.close()
        for index in self.indexes:
            index.close()
        self.env.close()

    def txn(self):
        return Transaction(self)

    @contextmanager
    def transaction(self):
        txn = self.txn()
        try:
            yield txn
        except Exception as exc:
            txn.rollback()
            raise exc
        else:
            txn.commit()
