#!/usr/bin/env python3
from traceback import format_exc
from datetime import datetime
from hashlib import md5

from .index import Index
from .index import IndexDescriptor


VERTEX, EDGE = range(2)


class MockStorage(dict):

    def put(self, key, value):
        self[key] = value

    def delete(self, key):
        del self[key]

    def get(self, key):
        return self.get(key, None)


class Element:

    def identifier(self):
        return self._identifier

    def label(self):
        return self._label

    def properties(self):
        return self._properties

    def __setitem__(self, key, value):
        try:
            self._properties[key]
        except KeyError:
            self._properties[key] = value
        else:
            try:
                self._previous_properties[key]
            except KeyError:
                previous_value = self._properties[key]
                self._previous_properties[key] = previous_value
            finally:
                self._properties[key] = value

    def __getitem__(self, key):
        return self._properties[key]

    def get(self, key, default):
        return self._properties.get(key, default)

    def old(self, key):
        return self._previous_properties[key]


class Vertex(Element):

    def __init__(self, txn, identifier, label, outgoings, incomings, properties):
        self._txn = txn
        self._identifier = identifier
        self._label = label
        self._outgoings = outgoings
        self._incomings = incomings
        self._properties = properties
        self._previous_properties = dict()

    @classmethod
    def create(cls, txn, label):
        # values is built as follow:
        # element type, label, outgoings, incomings, properties
        values = [VERTEX, label, list(), list(), dict()]
        # retrieve an identifier
        identifier = txn._random_identifier()
        # instantiate related object
        vertex = Vertex(txn, identifier, *values[1:])
        txn._cache[identifier] = vertex
        return vertex

    def _iter_over_edges(self, kind):
        kind = '_%s' % kind
        kind = getattr(self, kind, list())
        for identifier in kind:
            yield self._txn.get(identifier)

    def incomings(self):
        return self._iter_over_edges('incomings')

    def outgoings(self):
        return self._iter_over_edges('outgoings')
            

class Edge(Element):

    def __init__(self, txn, identifier, start, label, end, properties):
        self._txn = txn
        self._identifier = identifier
        self._start = start
        self._label = label
        self._end = end
        self._properties = properties
        self._previous_properties = dict()

    @classmethod
    def create(cls, txn, start, label, end):
        # see also comments in Vertex.create
        values = [
            EDGE, 
            start._identifier, 
            label, 
            end._identifier, 
            dict()
        ]
        identifier = txn._random_identifier()
        edge = Edge(txn, identifier, *values[1:])
        txn._cache[identifier] = edge
        # update start and end
        start._outgoings.append(edge.identifier())
        end._incomings.append(edge.identifier())
        return edge

    def start(self):
        return self._txn.get(self._start)

    def end(self):
        return self._txn.get(self._end)

    def label(self):
        return self._label


class MockTransaction:

    def __init__(self):
        self._dirty = set()
        self._cache = dict()

    def _random_identifier(self):
        while True:
            now = datetime.now()
            now = now.isoformat()
            identifier = md5(now.encode('ascii'))
            identifier = identifier.hexdigest()
            if self.get(identifier):
                continue
            else:
                break
        return identifier

    def query(self):
        query = Query(self)
        self._dirty.add(query)
        return query

    def all(self):
        yield from self._cache.values()

    def get(self, identifier):
        return self._cache.get(identifier, None)

    def vertex(self, label):
        return Vertex.create(self, label)

    def edge(self, start, label, end):
        return Edge.create(self, start, label, end)


tests = list()


def test(func):
    tests.append(func)


@test
def index_fresh_vertex():
    # init test
    storage = MockStorage()
    transaction = MockTransaction()
    # create artifical descroptor
    descriptor = IndexDescriptor('users').label('user').username.unique()
    # create index based on descriptor data using mock storage
    index = Index(descriptor._root(), storage)
    # populate transaction cache with data
    amirouche = transaction.vertex('user')
    amirouche['username'] = 'amirouche'
    index._update(amirouche)
    hocine = transaction.vertex('user')
    hocine['username'] = 'hocine'
    index._update(hocine)
    # those MUST not be indexed
    azoufzouf = transaction.vertex('animal')
    azoufzouf['username'] = 'azoufzouf'
    index._update(azoufzouf)
    nobody = transaction.vertex('user')
    index._update(nobody)
    assert len(storage) == 2


@test
def re_index_vertex():
    # init test
    storage = MockStorage()
    transaction = MockTransaction()
    # create artifical descriptor
    descriptor = IndexDescriptor('users').label('user').username.unique()
    # create index based on descriptor data using mock storage
    index = Index(descriptor._root(), storage)
    # populate transaction cache with data
    amirouche = transaction.vertex('user')
    amirouche['username'] = 'amirouche'
    index._update(amirouche)
    amirouche['username'] = 'amirouche2'
    index._update(amirouche)
    # old index MUST be deleted
    assert len(storage) == 1


@test
def ordered_index():
    # init test
    storage = MockStorage()
    transaction = MockTransaction()
    # create artifical descriptor
    favorites = IndexDescriptor('favorites').index('favorites').both.label('likes')
    favorites = favorites.property('created_at').order_by('created_at')
    # create index based on descriptor data using mock storage
    index = Index(favorites._root(), storage)
    # init favorites storage

    # populate transaction cache with data
    amirouche = transaction.vertex('user')
    amirouche['username'] = 'amirouche'
    index._update(amirouche)
    amirouche['username'] = 'amirouche2'
    index._update(amirouche)
    # old index MUST be deleted
    assert len(storage) == 1


if __name__ == '__main__':
    for func in tests:
        try:
            func()
        except Exception as e:
            print('-' * 70)
            print('%s: FAILED :(\n' % func.__name__)
            print(format_exc())
            print('-' * 70)
        else:
            print('%s: PASSED :)' % func.__name__)
