#!/usr/bin/env python3
from traceback import format_exc
from datetime import datetime
from hashlib import md5

import shutil

from .query import Query


VERTEX, EDGE = range(2)


class Element:

    def identifier(self):
        return self._identifier

    def label(self):
        return self._label

    def properties(self):
        return self._properties

    def __setitem__(self, key, value):
        self._properties[key] = value

    def __getitem__(self, key):
        return self._properties[key]

    def get(self, key, default):
        return self._properties.get(key, default)


class Vertex(Element):

    def __init__(self, txn, identifier, label, outgoings, incomings, properties):
        self._txn = txn
        self._identifier = identifier
        self._label = label
        self._outgoings = outgoings
        self._incomings = incomings
        self._properties = properties

    @classmethod
    def create(cls, txn, label):
        # values is built as follow:
        # element type, label, outgoings, incomings, properties
        values = [VERTEX, label, list(), list(), dict()]
        # retrieve an identifier
        identifier = txn._random_identifier()
        # instantiate related object
        vertex = Vertex(txn, identifier, *values[1:])
        txn._cache[identifier] = vertex
        return vertex

    def _iter_over_edges(self, kind):
        kind = '_%s' % kind
        kind = getattr(self, kind, list())
        for identifier in kind:
            yield self._txn.get(identifier)

    def incomings(self):
        return self._iter_over_edges('incomings')

    def outgoings(self):
        return self._iter_over_edges('outgoings')
            

class Edge(Element):

    def __init__(self, txn, identifier, start, label, end, properties):
        self._txn = txn
        self._identifier = identifier
        self._start = start
        self._label = label
        self._end = end
        self._properties = properties

    @classmethod
    def create(cls, txn, start, label, end):
        # see also comments in Vertex.create
        values = [
            EDGE, 
            start._identifier, 
            label, 
            end._identifier, 
            dict()
        ]
        identifier = txn._random_identifier()
        edge = Edge(txn, identifier, *values[1:])
        txn._cache[identifier] = edge
        # update start and end
        start._outgoings.append(edge.identifier())
        end._incomings.append(edge.identifier())
        return edge

    def start(self):
        return self._txn.get(self._start)

    def end(self):
        return self._txn.get(self._end)

    def label(self):
        return self._label


class MockTransaction:

    def __init__(self):
        self._dirty = set()
        self._cache = dict()

    def _random_identifier(self):
        while True:
            now = datetime.now()
            now = now.isoformat()
            identifier = md5(now.encode('ascii'))
            identifier = identifier.hexdigest()
            if self.get(identifier):
                continue
            else:
                break
        return identifier

    def query(self):
        query = Query(self)
        self._dirty.add(query)
        return query

    def all(self):
        yield from self._cache.values()

    def get(self, identifier):
        return self._cache.get(identifier, None)

    def vertex(self, label):
        return Vertex.create(self, label)

    def edge(self, start, label, end):
        return Edge.create(self, start, label, end)


tests = list()


def test(func):
    tests.append(func)


# Label tests
# - iterate
# - all
# - one

@test
def iterate_over_users_by_label_filtering():
    transaction = MockTransaction()
    amirouche = transaction.vertex('user')
    amirouche['name'] = 'amirouche'
    ahcene = transaction.vertex('user')
    ahcene['name'] = 'ahcene'
    hocine = transaction.vertex('user')
    hocine['name'] = 'hocine'
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    users = transaction.query().label('user')
    users = list(users)
    count = len(users)
    assert count == 3
    users = list(map(lambda x: x['name'], users))
    for name in ('amirouche', 'ahcene', 'hocine'):
        assert name in users


@test
def all_users_by_label_filtering():
    # same as above except the results isn't a generator
    transaction = MockTransaction()
    amirouche = transaction.vertex('user')
    amirouche['name'] = 'amirouche'
    ahcene = transaction.vertex('user')
    ahcene['name'] = 'ahcene'
    hocine = transaction.vertex('user')
    hocine['name'] = 'hocine'
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    # no need  for list(users)
    users = transaction.query().label('user').all()
    count = len(users)
    assert count == 3
    users = list(map(lambda x: x['name'], users))
    for name in ('amirouche', 'ahcene', 'hocine'):
        assert name in users


@test
def query_one_user_filtered_with_label():
    transaction = MockTransaction()
    amirouche = transaction.vertex('user')
    amirouche['name'] = 'amirouche'
    ahcene = transaction.vertex('user')
    ahcene['name'] = 'ahcene'
    hocine = transaction.vertex('user')
    hocine['name'] = 'hocine'
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    transaction.vertex('animal')
    user = transaction.query().label('user').one()
    assert user.label() == 'user'
    assert user['name'] in ('amirouche', 'ahcene', 'hocine')

# Property filtering tests
# - iterate
# - all
# - one

@test
def iterate_over_users_filtered_by_one_test():
    transaction = MockTransaction()
    amirouche = transaction.vertex('user')
    amirouche['name'] = 'amirouche'
    amirouche['age'] = 30
    ahcene = transaction.vertex('user')
    ahcene['name'] = 'ahcene'
    ahcene['age'] = 28
    hocine = transaction.vertex('user')
    hocine['name'] = 'hocine'
    hocine['age'] = 28
    mehand = transaction.vertex('user')
    mehand['name'] = 'mehand'
    mehand['age'] = 22
    users = transaction.query().filter(lambda x: x['age'] > 25)
    users = list(users)
    count = len(users)
    assert count == 3
    users = list(map(lambda x: x['name'], users))
    for name in ('amirouche', 'ahcene', 'hocine'):
        assert name in users


@test
def all_users_filtered_with_one_test():
    # same as above except the results isn't a generator
    transaction = MockTransaction()
    amirouche = transaction.vertex('user')
    amirouche['name'] = 'amirouche'
    amirouche['age'] = 30
    ahcene = transaction.vertex('user')
    ahcene['name'] = 'ahcene'
    ahcene['age'] = 28
    hocine = transaction.vertex('user')
    hocine['name'] = 'hocine'
    hocine['age'] = 28
    mehand = transaction.vertex('user')
    mehand['name'] = 'mehand'
    mehand['age'] = 22
    users = transaction.query().filter(lambda x: x['age'] > 25).all()
    count = len(users)
    assert count == 3
    for user in users:
        assert user['name'] in ('amirouche', 'ahcene', 'hocine')
        assert user['age'] > 25


@test
def query_one_user_filtered_with_one_test():
    transaction = MockTransaction()
    amirouche = transaction.vertex('user')
    amirouche['name'] = 'amirouche'
    amirouche['age'] = 30
    ahcene = transaction.vertex('user')
    ahcene['name'] = 'ahcene'
    ahcene['age'] = 28
    hocine = transaction.vertex('user')
    hocine['name'] = 'hocine'
    hocine['age'] = 28
    mehand = transaction.vertex('user')
    mehand['name'] = 'mehand'
    mehand['age'] = 22
    user = transaction.query().filter(lambda x: x['age'] > 25).one()
    assert user.label() == 'user'
    assert user['name'] in ('amirouche', 'ahcene', 'hocine')
    assert user['age'] > 25


@test
def iterate_over_users_filtered_with_two_tests():
    transaction = MockTransaction()
    amirouche = transaction.vertex('user')
    amirouche['name'] = 'amirouche'
    amirouche['age'] = 30
    ahcene = transaction.vertex('user')
    ahcene['name'] = 'ahcene'
    ahcene['age'] = 28
    hocine = transaction.vertex('user')
    hocine['name'] = 'hocine'
    hocine['age'] = 28
    mehand = transaction.vertex('user')
    mehand['name'] = 'mehand'
    mehand['age'] = 22
    users = transaction.query().filter(lambda x: x['age'] > 25)
    users = users.filter(lambda x: 'a' in x['name'])
    users = list(users)
    assert len(users) == 2
    for user in users:
        assert user['name'] in ('amirouche', 'ahcene')
        assert user['age'] > 25


@test
def iterate_over_users_filtered_with_two_tests_and_label():
    transaction = MockTransaction()
    amirouche = transaction.vertex('user')
    amirouche['name'] = 'amirouche'
    amirouche['age'] = 30
    ahcene = transaction.vertex('user')
    ahcene['name'] = 'ahcene'
    ahcene['age'] = 28
    hocine = transaction.vertex('user')
    hocine['name'] = 'hocine'
    hocine['age'] = 28
    mehand = transaction.vertex('user')
    mehand['name'] = 'mehand'
    mehand['age'] = 22
    animal = transaction.vertex('animal')
    animal['name'] = 'jeha'
    animal['age'] = 33
    users = transaction.query().label('user')
    users = users.filter(lambda x: x['age'] > 25)
    users = users.filter(lambda x: 'a' in x['name'])
    users = list(users)
    assert len(users) == 2
    for user in users:
        assert user['name'] in ('amirouche', 'ahcene')
        assert user['age'] > 25



@test
def TEMPLATE():
    transaction = MockTransaction()


if __name__ == '__main__':
    for func in tests:
        try:
            func()
        except Exception as e:
            print('%s: FAILED :(' % func.__name__)
            print(format_exc())
            print('-' * 80)
        else:
            print('%s: PASSED :)' % func.__name__)

